/*
  ## Cypress FX3 Camera Kit source file (sensor.c)
  ## ===========================
  ##
  ##  Copyright Cypress Semiconductor Corporation, 2010-2012,
  ##  All Rights Reserved
  ##  UNPUBLISHED, LICENSED SOFTWARE.
  ##
  ##  CONFIDENTIAL AND PROPRIETARY INFORMATION
  ##  WHICH IS THE PROPERTY OF CYPRESS.
  ##
  ##  Use of this file is governed
  ##  by the license agreement included in the file
  ##
  ##     <install>/license/license.txt
  ##
  ##  where <install> is the Cypress software
  ##  installation root directory path.
  ##
  ## ===========================
*/


/*
 * This file implements the I2C based driver for SerDes (MAX9705/MAX9706) and
 * Onsemi AP0202/AR0231 camera sensor. 
 */

#include <cyu3system.h>
#include <cyu3os.h>
#include <cyu3dma.h>
#include <cyu3error.h>
#include <cyu3uart.h>
#include <cyu3i2c.h>
#include <cyu3types.h>
#include <cyu3gpio.h>
#include <cyu3utils.h>
#include "sensor.h"

#define I2C_DEBUG

/* Write to an I2C slave with one byte of data. */
int
I2cWrite8(uint8_t slaveId, uint8_t addr, uint8_t data)
{
  CyU3PReturnStatus_t apiRetStatus = CY_U3P_SUCCESS;
  CyU3PI2cPreamble_t  preamble;
  uint8_t buf;

  /* Set the parameters for the I2C API access and then call the write API. */
  preamble.buffer[0] = slaveId;
  preamble.buffer[1] = addr;
  preamble.length    = 2;             /*  Three byte preamble. */
  preamble.ctrlMask  = 0x0000;        /*  No additional start and stop bits. */

  buf = data;

  apiRetStatus = CyU3PI2cTransmitBytes(&preamble, &buf, 1, 0);
  CyU3PBusyWait(I2C_DELAY_US);

  if (apiRetStatus != CY_U3P_SUCCESS) {
    CyU3PDebugPrint(4, "I2cWrite8: failed, slaveId=0x%x, addr=0x%x, data=0x%x, ErrorCode=%d\r\n",
                    slaveId, addr, data, apiRetStatus);
    return -1;
  }

#ifdef I2C_DEBUG
  CyU3PDebugPrint(4, "I2cWrite8: OK, slaveId=0x%x, addr=0x%x, data=0x%x\r\n",
                  slaveId, addr, data);
#endif

  return 0;
}

int
I2cRead8(uint8_t slaveId, uint8_t addr, uint8_t *datap)
{
  CyU3PReturnStatus_t apiRetStatus = CY_U3P_SUCCESS;
  CyU3PI2cPreamble_t preamble;

  preamble.buffer[0] = slaveId & I2C_SLAVEADDR_MASK;  /*  Mask out the transfer type bit. */
  preamble.buffer[1] = addr;
  preamble.buffer[2] = slaveId;
  preamble.length    = 3;
  preamble.ctrlMask  = 0x0002;    /* Send start bit after second byte of preamble. */

  apiRetStatus = CyU3PI2cReceiveBytes(&preamble, datap, 1, 0);
  CyU3PBusyWait(I2C_DELAY_US);

  if (apiRetStatus != CY_U3P_SUCCESS) {
    CyU3PDebugPrint(4, "I2cRead8: failed, slaveId=0x%x, addr=0x%x, ErrorCode=%d\r\n",
                    slaveId, addr, apiRetStatus);
    return -1;
  }

#ifdef I2C_DEBUG
  CyU3PDebugPrint(4, "I2cRead8: OK, slaveId=0x%x, addr=0x%x, data=0x%x\r\n",
                  slaveId, addr, *datap);
#endif
  return 0;
}

/* Write to an I2C slave with two bytes of data. */
int
I2cWrite16(uint8_t slaveId, uint16_t addr, uint16_t data)
{
  CyU3PReturnStatus_t apiRetStatus = CY_U3P_SUCCESS;
  CyU3PI2cPreamble_t  preamble;
  uint8_t buf[2];

  /* Set the parameters for the I2C API access and then call the write API. */
  preamble.buffer[0] = slaveId;
  preamble.buffer[1] = (addr & 0xff00) >> 8; /* high address */
  preamble.buffer[2] = addr & 0x00ff; /* low address */
  preamble.length    = 3;             /*  Three byte preamble. */
  preamble.ctrlMask  = 0x0000;        /*  No additional start and stop bits. */

  buf[0] = (data & 0xff00) >> 8;
  buf[1] = data & 0x00ff;

  apiRetStatus = CyU3PI2cTransmitBytes(&preamble, buf, 2, 0);
  CyU3PBusyWait(I2C_DELAY_US);

  if (apiRetStatus != CY_U3P_SUCCESS) {
    CyU3PDebugPrint(4, "I2cWrite16: failed, slaveId=0x%x, addr=0x%x, data=0x%x, ErrorCode=%d\r\n",
                    slaveId, addr, data, apiRetStatus);
    return -1;
  }

#ifdef I2C_DEBUG
  CyU3PDebugPrint(4, "I2cWrite16: OK, slaveId=0x%x, addr=0x%x, data=0x%x\r\n",
                  slaveId, addr, data);
#endif
  return 0;
}

int
I2cRead16(uint8_t slaveId, uint16_t addr, uint16_t *datap)
{
  CyU3PReturnStatus_t apiRetStatus = CY_U3P_SUCCESS;
  CyU3PI2cPreamble_t preamble;
  uint8_t buf[2];

  preamble.buffer[0] = slaveId & I2C_SLAVEADDR_MASK;  /*  Mask out the transfer type bit. */
  preamble.buffer[1] = (addr & 0xff00) >> 8;
  preamble.buffer[2] = addr & 0x00ff;
  preamble.buffer[3] = slaveId;
  preamble.length    = 4;
  preamble.ctrlMask  = 0x0004;    /* Send start bit after third byte of preamble. */

  apiRetStatus = CyU3PI2cReceiveBytes(&preamble, buf, 2, 0);
  *datap = buf[0] << 8 | buf[1];
  CyU3PBusyWait(I2C_DELAY_US);

  if (apiRetStatus != CY_U3P_SUCCESS) {
    CyU3PDebugPrint(4, "I2cRead16: failed, slaveId=0x%x, addr=0x%x, ErrorCode=%d\r\n",
                    slaveId, addr, apiRetStatus);
    return -1;
  }

#ifdef I2C_DEBUG
  CyU3PDebugPrint(4, "I2cRead16: OK, slaveId=0x%x, addr=0x%x, data=0x%x\r\n",
                  slaveId, addr, *datap);
#endif
  return 0;
}

int
I2cWrite16_data8(uint8_t slaveId, uint16_t addr, uint8_t data)
{
  CyU3PReturnStatus_t apiRetStatus = CY_U3P_SUCCESS;
  CyU3PI2cPreamble_t  preamble;
  uint8_t buf;

  /* Set the parameters for the I2C API access and then call the write API. */
  preamble.buffer[0] = slaveId;
  preamble.buffer[1] = (addr & 0xff00) >> 8;
  preamble.buffer[2] = addr & 0x00ff;
  preamble.length    = 3;             /*  Three byte preamble. */
  preamble.ctrlMask  = 0x0000;        /*  No additional start and stop bits. */

  buf = data;

  apiRetStatus = CyU3PI2cTransmitBytes(&preamble, &buf, 1, 0);
  CyU3PBusyWait(I2C_DELAY_US);

  if (apiRetStatus != CY_U3P_SUCCESS) {
    CyU3PDebugPrint(4, "I2cWrite16_data8: failed, slaveId=0x%x, addr=0x%x, data=0x%x, ErrorCode=%d\r\n",
                    slaveId, addr, data, apiRetStatus);
    return -1;
  }

#ifdef I2C_DEBUG
  CyU3PDebugPrint(4, "I2cWrite16_data8: OK, slaveId=0x%x, addr=0x%x, data=0x%x\r\n",
                  slaveId, addr, data);
#endif

  return 0;
}

int
I2cRead16_data8(uint8_t slaveId, uint16_t addr, uint8_t *datap)
{
  CyU3PReturnStatus_t apiRetStatus = CY_U3P_SUCCESS;
  CyU3PI2cPreamble_t preamble;

  preamble.buffer[0] = slaveId & I2C_SLAVEADDR_MASK;  /*  Mask out the transfer type bit. */
  preamble.buffer[1] = (addr & 0xff00) >> 8;
  preamble.buffer[2] = addr & 0x00ff;
  preamble.buffer[3] = slaveId;
  preamble.length    = 4;
  preamble.ctrlMask  = 0x0004;    /* Send start bit after third byte of preamble. */

  apiRetStatus = CyU3PI2cReceiveBytes(&preamble, datap, 1, 0);
  CyU3PBusyWait(I2C_DELAY_US);

  if (apiRetStatus != CY_U3P_SUCCESS) {
    CyU3PDebugPrint(4, "I2cRead16_data8: failed, slaveId=0x%x, addr=0x%x, ErrorCode=%d\r\n",
                    slaveId, addr, apiRetStatus);
    return -1;
  }

#ifdef I2C_DEBUG
  CyU3PDebugPrint(4, "I2cRead16_data8: OK, slaveId=0x%x, addr=0x%x, data=0x%x\r\n",
                  slaveId, addr, *datap);
#endif
  return 0;
}

/*
 * read AR0231 sensor via AP0202 ISP
 */
int
SensorRead16(uint16_t addr, uint16_t *datap)
{
  int ret = -1;
  uint8_t data8 = 0;
  int retry = SENSOR_REG_WAIT_MS / SENSOR_REG_RETRY_STEP;

  if (I2cWrite16(ISP_ADDR_WR, 0xF006, addr) != 0)
    goto done;
  if (I2cWrite16_data8(ISP_ADDR_WR, 0xF005, 0x04) != 0)
    goto done;
  /* poll the register value till it reaches 0 which means the read has finished */
  do {
    CyU3PBusyWait(SENSOR_REG_RETRY_STEP * 1000);
    ret = I2cRead16_data8(ISP_ADDR_RD, 0xF005, &data8);
    if (ret != 0)
      goto done;
  } while (data8 != 0 && retry--);
  if (data8 != 0)
    goto done;

  /* read out the register value */
  ret = I2cRead16(ISP_ADDR_RD, 0xF008, datap);

done:
  if (ret == 0) {
#ifdef I2C_DEBUG
    CyU3PDebugPrint(4, "SensorRead16 OK: addr=0x%x, data=0x%x\r\n",
                    addr, *datap);
#endif
  } else {
    CyU3PDebugPrint(4, "SensorRead16 failed: addr=0x%x\r\n", addr);
  }

  return ret;
}

/*
 * write AR0231 sensor via AP0202 ISP
 */
int
SensorWrite16(uint16_t addr, uint16_t data)
{
  uint8_t data8 = 0xFF;
  int retry = SENSOR_REG_WAIT_MS / SENSOR_REG_RETRY_STEP;

  if (I2cWrite16(ISP_ADDR_WR, 0xF006, addr) != 0)
    goto done;
  if (I2cWrite16(ISP_ADDR_WR, 0xF008, data) != 0)
    goto done;
  if (I2cWrite16_data8(ISP_ADDR_WR, 0xF005, 0x05) != 0)
    goto done;
  /* poll the register value till it reaches 0 which means the write has finished */
  do {
    CyU3PBusyWait(SENSOR_REG_RETRY_STEP * 1000);
    I2cRead16_data8(ISP_ADDR_RD, 0xF005, &data8);
  } while (data8 != 0 && retry--);

done:
  if (data8 == 0) {
#ifdef I2C_DEBUG
    CyU3PDebugPrint(4, "SensorWrite16 OK: addr=0x%x, data=0x%x\r\n", addr, data);
#endif
    return 0;
  } else {
    CyU3PDebugPrint(4, "SensorWrite16 failed: addr=0x%x, data=0x%x\r\n", addr, data);
    return -1;
  }
}

/*
 * Reset AP0202/AR0231 using GPIO.
 */
void
SensorReset(void)
{
  CyU3PReturnStatus_t apiRetStatus;

  /* Drive the GPIO low to reset the sensor. */
  apiRetStatus = CyU3PGpioSetValue(SENSOR_RESET_GPIO, CyFalse);
  if (apiRetStatus != CY_U3P_SUCCESS)
  {
    CyU3PDebugPrint(4, "SensorReset: GPIO Set Value 'False' failed, ErrorCode=%d\r\n", apiRetStatus);
    return;
  } 

  /* Wait for some time to allow proper reset. */
  CyU3PThreadSleep (10);

  /* Drive the GPIO high to bring the sensor out of reset. */
  apiRetStatus = CyU3PGpioSetValue(SENSOR_RESET_GPIO, CyTrue);
  if (apiRetStatus != CY_U3P_SUCCESS)
  {
    CyU3PDebugPrint(4, "SensorReset: GPIO Set Value 'True' failed, ErrorCode=%d\r\n", apiRetStatus);
    return;
  }

  /* Delay the allow the sensor to power up. */
  CyU3PThreadSleep(10);
  return;
}

void
SensorInit(void)
{
  int i = 0;
  int x = 0;
  uint8_t data8 = 0;
  uint16_t data16 = 0;

#ifdef USE_CAM10_CAM
  int n = 6;
  uint8_t ID[6] = {0x80, 0x80, 0x80, 0x80, 0x94, 0x94};
  uint8_t ADDR[6] = {0x0D, 0x0F, 0x01, 0x42, 0x06, 0x04};
  uint8_t DATA[6] = {0xB6, 0x80, 0x94, 0x5F, 0xEF, 0x87};

  wait_ready :

  // Wait 96706 ready
  for (i = 0; i < 100; i++) {
    I2cRead8(DES_ADDR_RD, 0x01, &data8);
    if (data8 == DES_ADDR_WR) {
      CyU3PThreadSleep(10);
      I2cWrite8(DES_ADDR_WR, 0x0D, 0xB6);
      CyU3PThreadSleep(10);
      I2cWrite8(DES_ADDR_WR, 0x04, 0x47);
      CyU3PThreadSleep(10);
      I2cWrite8(SER_ADDR_WR, 0x07, 0x84);
      CyU3PThreadSleep(10);
      I2cWrite8(DES_ADDR_WR, 0x07, 0x86);
      break;
    } else {
      CyU3PThreadSleep(10);
    }
  }

  CyU3PThreadSleep(10);

  // Wait 96705 ready
  for (i = 0; i < 100; i++) {
    I2cRead8(SER_ADDR_RD, 0x00, &data8);
    if (data8 != SER_ADDR_WR) {
      CyU3PThreadSleep(10);
      I2cWrite8(DES_ADDR_WR, 0x0D, 0xB6);
      CyU3PThreadSleep(10);
      I2cWrite8(DES_ADDR_WR, 0x04, 0x47);
      CyU3PThreadSleep(10);
      I2cWrite8(SER_ADDR_WR, 0x07, 0x84);
      CyU3PThreadSleep(10);
      I2cWrite8(DES_ADDR_WR, 0x07, 0x86);
      CyU3PThreadSleep(10);
    } else {
      break;
    }
  }

  // stream on
  for (x = 0; x < n; x++) {
    for (i = 0; i < 100; i++) {
      I2cRead8(ID[x]+1, ADDR[x], &data8);
      if (data8 != DATA[x]) {
        I2cWrite8(ID[x], ADDR[x], DATA[x]);
      } else {
        break;
      }
      CyU3PThreadSleep(10);
    }
  }
  I2cRead16(ISP_ADDR_RD, 0xC89A, &data16);
  CyU3PBusyWait (10);
  if (data16 != 0x4564) {  // check if it's CAM10
    goto wait_ready;
  }

#else
  // init Serdes link
  I2cWrite8(DES_ADDR_WR, 0x0D, 0xB6);
  I2cWrite8(SER_ADDR_WR, 0x01, 0x94);
  I2cWrite8(SER_ADDR_WR, 0x07, 0x84);
  I2cWrite8(DES_ADDR_WR, 0x07, 0x86);
  I2cWrite8(SER_ADDR_WR, 0x0D, 0xB6);
  I2cWrite8(SER_ADDR_WR, 0x0D, 0xD9);
  I2cWrite8(DES_ADDR_WR, 0x0D, 0xD9);
  I2cWrite8(SER_ADDR_WR, 0x0F, 0x81);  // enable trigger

  // init sensor image config: YUV
  I2cWrite16(ISP_ADDR_WR, 0xC98E, 0x4A9C);
  I2cWrite16(ISP_ADDR_WR, 0xCAFC, 0x4A09);
  I2cWrite16(ISP_ADDR_WR, 0xFC00, 0x2800);
  I2cWrite16(ISP_ADDR_WR, 0x0040, 0x8100);
  I2cWrite16(ISP_ADDR_WR, 0x0040, 0x8101);
#endif  // USE_CAM10_CAM
}
